from sklearn.neighbors import BallTree
import xml.etree.ElementTree as ET
# from gcn_kafka import Consumer
import numpy as np
from skoal.paths import TESS_DIR
from skoal.GCN_utils import getFERMICoordinates

# please just use this coordinate changer as others, such as astropy's, do
# not take/return the correct format 
def spherical_to_cartesian(spherical_cartesian_coords):
    theta = np.radians(spherical_cartesian_coords[:, 0])
    phi = np.radians(spherical_cartesian_coords[:, 1] +90)
    x = np.sin(phi) * np.cos(theta)
    y = np.sin(phi) * np.sin(theta) 
    z = np.cos(phi)
    cartesian_coords = np.column_stack((x, y, z))
    return cartesian_coords

def Fermi_handle(telescope, eventfile, rafov, decfov):
    # Find the C1, C2, and Error2Radius elements
    error_buff = (((rafov**2) + (decfov**2))**(1/2))/2
    ra, dec, error = getFERMICoordinates(eventfile)
    buffed_error = error + error_buff
    #get everything in radians
    ra = np.radians(ra)
    dec= np.radians(dec+90)
    center = [[np.sin(dec) * np.cos(ra),np.sin(dec) * np.sin(ra),np.cos(dec)]]
    #get cartesian error radius
    r = 2*np.sin(np.radians(buffed_error)/2)
    
    # Read the file and extract the second and third columns
    data = np.loadtxt(f'{TESS_DIR}/{telescope}.tess', usecols=(1, 2))
    # Create BallTree
    Tree = BallTree(spherical_to_cartesian(data), leaf_size=40)

    field_ids = list(Tree.query_radius(center, r=r,sort_results=True,return_distance=True)[0][0])
    pointers = [list(data[index]) for index in field_ids]

    # Write targets to file
    # with open(outpath, 'w') as file:
    #     for field, pointer in zip(field_ids, pointers):
    #         file.write(f"{field},{pointer[0]},{pointer[1]},1\n")
    # print(pointers)
    targets = [(field_ids[i-1], pointers[i-1][0], pointers[i-1][1]) for i in range(len(field_ids))]
    return targets, error



import numpy as np
import astropy.units as u
from astropy.coordinates import SkyCoord, EarthLocation, AltAz
from astropy.time import Time
from pathlib import Path

DIR = Path(__file__).parent.absolute()


def determine_visibility(input_file, observer_location, output_file, observation_time=None):
    tstart = Time.now()
    empty = False
    # Load input file
    data = np.genfromtxt(input_file, delimiter=',', dtype=str)
    ras = data[:, 1].astype(float) * u.deg
    decs = data[:, 2].astype(float) * u.deg
    priorities = data[:, 3].astype(int)

    # Convert input coordinates to SkyCoord objects
    coords = SkyCoord(ra=ras, dec=decs, frame='icrs')

    # Define observation time (default to current time if not provided)
    if observation_time is None:
        observation_time = Time.now() #+ 6.*u.hour

    # Convert observation time to AltAz frame
    altaz_frame = AltAz(obstime=observation_time, location=observer_location)
    coords_altaz = coords.transform_to(altaz_frame)

    # Check altitude of each object
    observable_indices = np.where(coords_altaz.alt > 0)[0]
    if len(observable_indices) ==0:
        empty = True
    time= Time.now()-tstart
    print(time)
    # Write observable objects to output file
    with open(output_file, 'w') as f:
        for idx in observable_indices:
            f.write(f"{data[idx, 0]},{data[idx, 1]},{data[idx, 2]},{data[idx, 3]}\n")
    return empty


def save_table_to_file(table, filename):
    np.savetxt(filename, table, delimiter=',', fmt='%s')


def split_schedule(infile, outfolder, num_tables):
    tables = [[] for _ in range(num_tables)]
    data=np.genfromtxt(infile,delimiter=',',dtype=str)
    
    for i, row in enumerate(data):
        table_index = i % num_tables
        tables[table_index].append(row)
    for i, table in enumerate(tables):
        table_filename = f"{outfolder}/telescope_{i+1}.txt"  # Name of the text file
        save_table_to_file(table, table_filename)
        #print(f"Schedule {i+1} saved to '{table_filename}'")

    return tables
# loc = EarthLocation(lat=44.990307*u.deg, lon=-93.179719*u.deg, height=242*u.m)

# determine_visibility(f'{DIR}/current_schedules/CurrentEventSchedule.txt', loc,f'{DIR}/current_schedules/ObservableCurrentEventSchedule.txt' )



from astroplan import FixedTarget, AirmassConstraint, AtNightConstraint, MoonSeparationConstraint, AltitudeConstraint, is_observable, Observer
from scipy.optimize import linear_sum_assignment
from scipy.spatial.distance import cdist
from astropy.coordinates import SkyCoord, EarthLocation
from astropy.time import Time
import astropy.units as u
from sklearn.cluster import KMeans
import numpy as np

def read_targets_from_file(file_name):
    targets = []
    for row in np.genfromtxt(file_name, delimiter=',', dtype=str):
        # (ra, dec, name, exposure)
        targets.append([float(row[1]), float(row[2]), row[0], 30])
    return targets

def _get_even_clusters(X, n_clusters):
    # Source: https://stackoverflow.com/questions/5452576/k-means-algorithm-variation-with-equal-cluster-size
    cluster_size = int(np.ceil(len(X)/n_clusters))
    kmeans = KMeans(n_clusters, n_init='auto')
    kmeans.fit(X)
    centers = kmeans.cluster_centers_
    centers = centers.reshape(-1, 1, X.shape[-1]).repeat(
        cluster_size, 1).reshape(-1, X.shape[-1])
    distance_matrix = cdist(X, centers)
    clusters = linear_sum_assignment(distance_matrix)[1]//cluster_size
    return clusters


def separate_targets_into_clusters(targets, n):
    # Convert targets into 3D points on the unit sphere
    targets_3d = []
    for target in targets:
        targets_3d.append([np.cos(np.radians(target[1]))*np.cos(np.radians(target[0])),
                           np.cos(np.radians(target[1]))*np.sin(np.radians(target[0])),
                           np.sin(np.radians(target[1]))])
    # Perform k-means clustering
    clusters = _get_even_clusters(np.array(targets_3d), n)
    # Split blocks between the telescopes.
    telescope_targets = [[] for _ in range(n)]
    for (label, target) in zip(clusters, targets):
        telescope_targets[label].append(target)

    return telescope_targets

def separate_targets_evenly(targets, n):
    # Split blocks between the telescopes.
    telescope_targets = [[] for _ in range(n)]
    for i in range(len(targets)):
        telescope_targets[i % n].append(targets[i])

    return telescope_targets

def filter_for_visibility(targets, lat, lon, altitude, twilight, telescope, horizon):

    # Constraints for the telescope
    constraints = [AirmassConstraint(max=2, boolean_constraint=False),
                    # AtNightConstraint.twilight_nautical() 
                    # if twilight == 'nautical' else AtNightConstraint.twilight_astronomical(),
        MoonSeparationConstraint(min=5*u.deg),
        AltitudeConstraint(min=horizon*u.deg)]
    loc = EarthLocation(lat=float(lat)*u.deg, 
                        lon=float(lon)*u.deg, 
                        height=float(altitude)*u.m)
    observer = Observer(loc, name=telescope, timezone="US/Central")    
    # The offset is for testing what happens if you schedule during the night
    t = Time.now()+12.*u.hour
    t_start = Time(t, format='jd')

    # Change targets to FixedTarget's
    fixed_targets = []
    for target in targets:
        fixed_targets.append(FixedTarget(coord=SkyCoord(
            target[1]*u.deg, target[2]*u.deg), name=target[0]))

    # Check if targets will be observable in the time window
    observable = is_observable(constraints, observer, fixed_targets, times=[t_start])
    # print(observable)
    filtered_targets = []
    for i in range(len(targets)):
        if observable[i]:
            filtered_targets.append(targets[i])
    return filtered_targets


import argparse
import configparser
import os
from pathlib import Path
from skoal.tesselation_generator import rect_tess_maker
import skoal.GCN_utils as gcn
from skoal.lvc_handler import generate_fields_from_skymap, save_targets_to_file
from skoal.Fermi_handler import Fermi_handle
from skoal.scheduler_utilities import filter_for_visibility
from skoal.config_utils import make_config_file
from skoal.Multiscope_handler import split_schedule
import numpy as np
from skoal.paths import SKOAL_DIR, CONFIGS_DIR, TESS_DIR, SKYMAPS_DIR
import time



def main():
    # tstart = time.time()
    parser = argparse.ArgumentParser(description="")
    parser.add_argument('-t', '-telescope', dest='telescope', type=str)
    parser.add_argument('-voe', '-voevent', dest='voevent', type=str)
    parser.add_argument('-e','-event', dest='event', type=str)
    parser.add_argument('-area','-a', action='store_true')
    parser.add_argument('-o','-outpath', dest='outpath', type=str)
    parser.add_argument('-multiscopes','-multiscope', '-m', dest='Number_of_telescopes', type=str)
    parser.add_argument('-alltargets','-alltargs','-alltarg', action='store_true')


    args = parser.parse_args()
    outpath = args.outpath
    telescope = args.telescope
    eventfile = args.voevent
    eventid = args.event
    multiscopes=args.Number_of_telescopes
    alltargets = args.alltargets
    
    
    if not eventfile and not eventid:
        print('no event/eventfile given...')
        entry = input('please provide event name or VOEvent.xml file: ')
        if entry[-4:] == '.xml':
            eventfile = entry
        else:
            eventid = entry
        if not entry:
            exit("No response, exiting...")
        #FERMI test
        # eventfile = f'{TESTS_DIR}/gcn.classic.voevent.FERMI_GBM_POS_TEST_4586.xml'

        #LVC test
        # eventfile = f'{TESTS_DIR}/gcn.classic.voevent.LVC_INITIAL_7496.xml'


    
    if eventfile:
        if eventid:
                print('it is unnecessary to provide both event and event file, either will do')
        if gcn.get_ivorn(eventfile) == 'Fermi':
            if eventid:
                print('if you are going to give both an event and event file, make sure theyre for the same type of notice')
                exit()
            print('FERMI eventfile given')
            fermi = True
            lvc = False
        elif gcn.get_ivorn(eventfile) == 'LVC':
            if eventid and not eventid == gcn.getEvent(eventfile):
                print('if you are going to give both an event and event file, make sure theyre for the same event')
                exit()
            eventid = gcn.getEvent(eventfile)
            print(f'LVC eventfile given, event {eventid} grabbed from eventfile')
            lvc = True
            fermi = False
        else:
            print('Notice type not supported')
            exit()      
    else:
        lvc = True
        fermi=False


    if telescope == None:
        print('telescope not given...')
        telescope = input('please enter telescope name: ')
        if not telescope:
            exit("No response, exiting...")
        # telescope = 'RASA11'

    config = configparser.ConfigParser()
    default = configparser.ConfigParser()
    default.read(f'{SKOAL_DIR}/data/configs/default.cfg')

    if os.path.exists(f'{CONFIGS_DIR}/{telescope}.cfg'):
        config.read(f'{CONFIGS_DIR}/{telescope}.cfg')
    else:
        make_config_file(telescope, CONFIGS_DIR)
        config.read(f'{CONFIGS_DIR}/{telescope}.cfg')

    try:
        lat = float(config[telescope]['lat'])
        lon = float(config[telescope]['lon'])
        RAfov = float(config[telescope]['RAfov'])
        DECfov = float(config[telescope]['DECfov'])
        elevation = float(config[telescope]['elevation'])
    except:
        print('flemopt config files MUST contain the following items: lat, lon, RAfov, DECfov and elevation. Default settings will be used for parameters not provided in configuration file')
        exit()
    
    
    try:
        tileScale = float(config[telescope]['tileScale'])
    except:
        print('tilescale not provided in .cfg file, using default setting')
        tileScale = float(default['telescope']['tileScale'])

    if lvc:
        try:
            minObsChance = float(config[telescope]['minObsChance'])
        except:
            minObsChance = float(default['telescope']['minObsChance'])
            print('minimum observation chance not provided in .cfg file, using default setting')


    try:
        horizon = float(config[telescope]['horizon'])
    except:
        horizon = float(default['telescope']['horizon'])
        print('horizon angle not provided in .cfg file, using default setting')


    if not os.path.exists(f'{TESS_DIR}/{telescope}.tess'):
        rect_tess_maker(telescope, RAfov, DECfov, tileScale)
    if not outpath:
        try:
            outpath = config[telescope]['default_output_path']
        except:
            outpath = default['telescope']['default_output_path']


    if not os.path.exists(outpath):
        os.mkdir(outpath)
        
    tstart = time.time()

    if lvc:
        skymap_path =gcn.get_skymap(eventid, SKYMAPS_DIR) 
        sorted_fields, ids_to_fields = generate_fields_from_skymap(skymap_path, RAfov, DECfov, tileScale, minObsChance)
        # print(sorted_fields)
        targets = [(id, np.rad2deg(ids_to_fields[id][0]), np.rad2deg(ids_to_fields[id][1])) for id, _ in sorted_fields]
        if args.area:
            gcn.area(minObsChance, skymap_path)
        outname = eventid
        # print(targets)
        # print(ids_to_fields)
    
    if fermi:
        targets, error = Fermi_handle(telescope, eventfile, RAfov, DECfov)
        # print(targets)
        if args.area:                     
            print(f'Search area is: {np.pi*(error^2)} square degrees')
        outname = 'FERMI_targets'

    
    print(f'field finding time: {time.time()-tstart}')
    
    if not alltargets:
        tstart = time.time()
        print(f'filtering {len(targets)} targets for observability')
        targets = filter_for_visibility(targets, lat, lon, elevation, 'nautical', telescope, horizon)
        print(f'time spent on observability filtering: {time.time()-tstart}')
        if len(targets) == 0:
            print('No observable targets')
            exit()
        else:
            print(f'found {len(targets)} observable targets')
    else:
        print(f'skipping observability filtering for {len(targets)} targets')


    

    if multiscopes and not int(multiscopes) == 1:
        try:
            os.mkdir(f'{outpath}/{outname}')
        except:
            print('output directory already exists, using existing path')
        try:
            outpath=f'{outpath}/{outname}'
            inpath=f'{outpath}/{telescope}_array_targets.txt'
            save_targets_to_file(targets, inpath)
            tables = split_schedule(inpath, outpath, int(multiscopes))
            print(f'successfully saved {multiscopes} lists with {len(tables[int(int(multiscopes)/2)])} targets each')
        except:
            print('failed to save targets successfully, exiting...')
            exit()
    else:
        try:
            save_targets_to_file(targets, f'{outpath}/{outname}.txt')
            print(f'successfully saved to {outpath}/{outname}')
        except:
            print('failed to save targets successfully, exiting...')
            exit()



if __name__ == '__main__':
    main()



import numpy as np
from sklearn.neighbors import BallTree
from astropy.table import Table
import xml.etree.ElementTree as ET
from pathlib import Path

wk_dir = Path(__file__).parent.absolute()

def rect_tess_maker(telescope, rafov,decfov, scale=0.97):
    rafov = np.deg2rad(rafov)*scale
    decfov = np.deg2rad(decfov)*scale
    vertical_count = np.ceil(np.pi / (decfov))  

    phis = []
    thetas = []

    phi = -0.5 * np.pi
    phi_step = np.pi / vertical_count
    # print('phi_step is', np.rad2deg(phi_step))
    while phi < 0.5 * np.pi - 1e-8 :
        if phi < (0-(phi_step/2)) and not phi == (-.5*np.pi):
            horizontal_count = np.ceil(
                (2 * np.pi * np.cos(phi + (phi_step/2))) / (rafov)
            )
        elif phi > (phi_step/2) and phi<((.5*np.pi)-(phi_step/2)):
            horizontal_count = np.ceil(
                (2 * np.pi * np.cos(phi - (phi_step/2))) / (rafov)
            )
        elif phi < (phi_step/2) and phi > (-phi_step/2):
            horizontal_count = np.ceil(2*np.pi/rafov)
        else:
            horizontal_count = 1
        # print(horizontal_count)
        theta = 0
        theta_step = 2 * np.pi / horizontal_count
        # print(np.rad2deg(theta_step))
        while theta < 2 * np.pi - 1e-8:
            phis.append(phi)
            thetas.append(theta)
            theta += theta_step
        phi += phi_step
    phis.append(np.pi/2)
    thetas.append(1)
    dec = np.array(np.rad2deg(phis))
    ra = np.array(np.rad2deg(thetas))
    tessfile_path = f'{wk_dir}/data/tesselations/{telescope}.tess'


    fid = open(tessfile_path, "w")
    for ii in range(len(ra)):
        fid.write("%d %.5f %.5f\n" % (ii, ra[ii], dec[ii]))
    fid.close()


# rect_tess_maker( 'kasa11', 3.2, 2.1, .98)






import numpy as np
import time

tstart = time.time()
def field_from_coords(coords, rafov,decfov, scale=.97):

    rafov *= scale
    decfov *= scale
    rafov = np.deg2rad(rafov)
    decfov = np.deg2rad(decfov)
    phis = np.ceil(np.pi/decfov)
    phi_step = np.pi /phis
    # print(phis)
    # print(np.rad2deg(phi_step))


    phi = -np.pi/2
    phis = []
    thetas = []
    thetasteps = []
    while phi < 0.5 * np.pi - 1e-8 :
        if phi < (0-(phi_step/2)) and not phi == (-.5*np.pi):
            horizontal_count = np.ceil(
                (2 * np.pi * np.cos(phi + (phi_step/2))) / (rafov)
            )
        elif phi > (phi_step/2) and phi<((.5*np.pi)-(phi_step/2)):
            horizontal_count = np.ceil(
                (2 * np.pi * np.cos(phi - (phi_step/2))) / (rafov)
            )
        elif phi < (phi_step/2) and phi > (-phi_step/2):
            horizontal_count = np.ceil(2*np.pi/rafov)
        else:
            horizontal_count = 1
        hc=int(horizontal_count)
        thetas.append(hc)
        thetasteps.append(2*np.pi/hc)
        phis.append(phi)
        phi += phi_step
    thetas.append(1)
    thetasteps.append(2*np.pi)
    phis.append(np.pi/2)
    # print(phis)

    start_index = []
    count = 0
    for n in thetas:
        start_index.append(count)
        count += n


    field_ids = []
    centers = []

    ddexs= []
    raNs=[]


    for ra,dec in coords:
        ddex = dec_num(dec,phi_step)
        # print(dec, phi_step)

        # print(thetas)
        raN = ra_number(ra, thetas[ddex]) 
        id = raN + start_index[ddex]
        field_ids.append(id-1)

        ddexs.append(ddex)
        raNs.append(raN)
        # print(raN, ddex)

        centers.append((thetasteps[ddex]*(raN-1), phis[ddex]))

    return field_ids, centers
        



def ra_number(ra, thetas):
    h = np.floor((ra/ (2*np.pi/thetas)) + (3/2))
    if h == thetas + 1:
        h = 1
    return int(h)

def dec_num(dec,phi_step):
    a = (dec+(np.pi/2))/phi_step
    num=np.round(a) 
    return int(num)


# coords = np.deg2rad(np.array([[0,-89.999], [185.0, 90.0]]))
# rafov = 10
# decfov = 5
# field_ids, centers = field_from_coords(coords, rafov, decfov)
# print(field_ids)
# print(centers)
# print(time.time()-tstart)


from pathlib import Path
import os



SKOAL_DIR = Path(__file__).parent.absolute()
DATA_DIR = SKOAL_DIR.joinpath("data")

CONFIGS_DIR = DATA_DIR.joinpath("configs")
TESS_DIR = DATA_DIR.joinpath('tesselations')
SKYMAPS_DIR = DATA_DIR.joinpath('skymaps')
TESTS_DIR = DATA_DIR.joinpath('test_eventfiles')

if not os.path.exists(DATA_DIR):
    os.mkdir(DATA_DIR)
if not os.path.exists(CONFIGS_DIR):
    os.mkdir(CONFIGS_DIR)
if not os.path.exists(TESS_DIR):
    os.mkdir(TESS_DIR)
if not os.path.exists(SKYMAPS_DIR):
    os.mkdir(SKYMAPS_DIR)






import configparser

def make_config_file(telescope, CONFIGS_DIR):
    config = configparser.ConfigParser()
    lat = input(f'please enter {telescope} latitude: ')
    lon = input(f'please enter {telescope} longitude: ')
    RAfov = input(f'please enter {telescope} RAfov: ')
    DECfov = input(f'please enter {telescope} DECfov: ')
    elevation = input(f'please enter {telescope} elevation: ')
    config.add_section(telescope)
    config.set(telescope, 'lat', lat)
    config.set(telescope, 'lon', lon)
    config.set(telescope, 'RAfov', RAfov)
    config.set(telescope, 'DECfov', DECfov)
    config.set(telescope, 'elevation', elevation)
    with open(f'{CONFIGS_DIR}/{telescope}.cfg', 'w') as configfile:
        config.write(configfile)




from astropy.table import QTable
import astropy_healpix as ah
from pathlib import Path
import numpy as np
from skoal.faster_fieldfinder import field_from_coords

FLEMOPT_DIR = Path(__file__).parent.absolute()

def generate_fields_from_skymap(skymap_path, rafov, decfov, scale, minobs):
    # Read the file
    
    skymap = QTable.read(skymap_path)

    # Sort by probability density
    skymap.sort('PROBDENSITY', reverse=True)
    # Get the pixel area
    level, nested_index = ah.uniq_to_level_ipix(skymap['UNIQ'])
    pixel_area = ah.nside_to_pixel_area(ah.level_to_nside(level))
    # Calculate the probability of each pixel
    probabilities = pixel_area * skymap['PROBDENSITY']
    # Find minimum cutoff
    cumprob = np.cumsum(probabilities)
    i = cumprob.searchsorted(minobs)

    # Find the corresponding tesselations of those pixels
    sky_coordinates = ah.healpix_to_lonlat(nested_index, ah.level_to_nside(level), order='nested')
    sky_coordinates = np.array(sky_coordinates).transpose()
    ids, fields = field_from_coords(sky_coordinates[:i], rafov, decfov, scale)
    ids_to_fields = dict(zip(ids, fields))

    # Sum the probabilities of the tesselations
    weights = {}
    for id, probability in zip(ids, probabilities):
        if id in weights:
            weights[id] += probability
        else:
            weights[id] = probability
    # Sort by total probability
    sorted_fields = sorted(weights.items(), key=lambda item: item[1], reverse=True)
    # print(sorted_fields)
    return sorted_fields, ids_to_fields

def save_targets_to_file(filtered_targets, out_dir):
    with open(out_dir, 'w') as file:
        for target in filtered_targets:
            id = target[0]
            ra_deg = target[1]
            dec_deg = target[2]
            file.write(f"{id},{ra_deg:.5f},{dec_deg:.5f},1\n")
    


import numpy as np


def field_from_coords(coords, rafov, decfov, scale=0.97):
    rafov *= scale
    decfov *= scale
    rafov = np.deg2rad(rafov)
    decfov = np.deg2rad(decfov)
    phis_count = int(np.ceil(np.pi / decfov))
    phi_step = np.pi / phis_count

    phis = []
    thetas = []
    thetasteps = []
    phi = -np.pi / 2

    while phi < 0.5 * np.pi - 1e-8:
        if phi < (0 - (phi_step / 2)) and phi != (-0.5 * np.pi):
            horizontal_count = np.ceil((2 * np.pi * np.cos(phi + (phi_step / 2))) / rafov)
        elif phi > (phi_step / 2) and phi < ((0.5 * np.pi) - (phi_step / 2)):
            horizontal_count = np.ceil((2 * np.pi * np.cos(phi - (phi_step / 2))) / rafov)
        elif phi < (phi_step / 2) and phi > (-phi_step / 2):
            horizontal_count = np.ceil(2 * np.pi / rafov)
        else:
            horizontal_count = 1
        hc = int(horizontal_count)
        thetas.append(hc)
        thetasteps.append(2 * np.pi / hc)
        phis.append(phi)
        phi += phi_step

    thetas.append(1)
    thetasteps.append(2 * np.pi)
    phis.append(np.pi / 2)

    start_index = np.zeros(len(thetas), dtype=np.int64)
    for i in range(1, len(thetas)):
        start_index[i] = start_index[i-1] + thetas[i-1]

    num_coords = len(coords)
    field_ids = np.empty(num_coords, dtype=np.int64)
    centers = np.empty((num_coords, 2), dtype=np.float64)

    for i in range(num_coords):
        ra, dec = coords[i]
        ddex = int(np.round((dec + np.pi / 2) / phi_step))
        ddex = max(0, min(ddex, len(phis) - 1))
        raN = int(np.floor((ra / (2 * np.pi / thetas[ddex])) + 1.5))
        if raN == (thetas[ddex] + 1):
            raN = 1
        id = raN + start_index[ddex] - 1
        field_ids[i] = id
        centers[i] = (thetasteps[ddex] * (raN - 1), phis[ddex])

    return field_ids.tolist(), centers.tolist()

# # Example usage
# coords = np.deg2rad(np.array([[0, -89.999], [185.0, 90.0]]))
# rafov = 10
# decfov = 5

# import time

# start = time.time()
# field_ids, centers = field_from_coords(coords, rafov, decfov)
# end = time.time()
# print(field_ids)
# print(centers)
# print(end - start)



from setuptools import setup, find_packages

setup(
    name='Skoal',
    version='0.422',
    packages=find_packages(),
    scripts=['skoal/main.py'],
    entry_points={
        'console_scripts': [
            'skoal=skoal.main:main',
        ],
    },
    package_data={
        # Include all files inside mypackage/data
        'skoal': ['data/**/*'],
    },
    install_requires=[
        'numpy',
        'scikit-learn',
        'astropy',
        'astropy-healpix',
        'astroplan',
        'scipy',
        'argparse',
        'configparser',
        'requests',
        'ligo-gracedb>=2.10.0',
        'lxml',
        'urllib3>=2.1.0',
    ],
    author='Benny Border',
    author_email='borderbenja@gmail.com',
    description='a kilonova followup scheduling package for fermi and lvc notices',
    long_description=open('README.md').read(),
    long_description_content_type='text/markdown',
    url='https://github.com/borderbenja05/skoal',
    classifiers=[
        'Programming Language :: Python :: 3',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
    ],
    python_requires='>=3.6',
)



import xml.etree.ElementTree as ET
import os
from urllib.parse import urlparse
import astropy.units as u
from astropy.coordinates import SkyCoord
import astropy_healpix as ah
from astropy.table import QTable
import numpy as np
import requests
from pathlib import Path
from ligo.gracedb.rest import GraceDb
import lxml.etree


def area(total_probability, bayestar_file):
    skymap = QTable.read(bayestar_file)
    skymap.sort('PROBDENSITY', reverse=True)
    level = ah.uniq_to_level_ipix(skymap['UNIQ'])
    pixel_area = ah.nside_to_pixel_area(ah.level_to_nside(level))
    prob = pixel_area * skymap['PROBDENSITY']
    cumprob = np.cumsum(prob)
    i = cumprob.searchsorted(total_probability)
    area_90 = pixel_area[:i].sum()
    area=area_90.to_value(u.deg**2)
    print(area)
    return area

def download_from_url(skymap_url: str, output_dir: Path, skymap_name: str) -> Path:
    """
    Download a skymap from a URL

    :param skymap_url: URL to download from
    :param output_dir: Output directory
    :param skymap_name: Name of skymap
    :return: Path to downloaded skymap
    """
    output_dir = Path(output_dir)
    savepath = output_dir.joinpath(skymap_name)

    if savepath.exists():
        print(f"File {savepath} already exists. Using this.")
    else:
        print(f"Saving to: {savepath}")
        response = requests.get(skymap_url, headers={"User-Agent": "Mozilla/5.0"})

        with open(savepath, "wb") as f:
            f.write(response.content)

    return savepath


def get_skymap_gracedb(
    event_name: str, rev=None, output_dir: Path = 'SKYMAP_DIR'
) -> Path:
    """
    Fetches the skymap from GraceDB

    :param event_name: name of the event
    :param rev: revision number of the event
    :param output_dir: directory to save the skymap and event info
    :return: path to the skymap
    """
    ligo_client = GraceDb()

    voevents = ligo_client.voevents(event_name).json()["voevents"]

    if rev is None:
        rev = len(voevents)

    elif rev > len(voevents):
        raise Exception(f"Revision {0} not found".format(rev))

    latest_voevent = voevents[rev - 1]
    print(f"Found voevent {latest_voevent['filename']}")

    if "Retraction" in latest_voevent["filename"]:
        raise ValueError(
            f"The specified LIGO event, "
            f"{latest_voevent['filename']}, was retracted."
        )

    response = requests.get(
        latest_voevent["links"]["file"],
        headers={"User-Agent": "Mozilla/5.0"},
        timeout=60,
    )

    root = lxml.etree.fromstring(response.content)
    params = {
        elem.attrib["name"]: elem.attrib["value"] for elem in root.iterfind(".//Param")
    }

    latest_skymap_url = params["skymap_fits"]

    print(f"Latest skymap URL: {latest_skymap_url}")

    skymap_name = "_".join(
        [event_name, str(latest_voevent["N"]), os.path.basename(latest_skymap_url)]
    )

    skymap_path = download_from_url(latest_skymap_url, output_dir, skymap_name)

    return skymap_path

    


def get_skymap(event_name: str, output_dir: Path = 'SKYMAP_DIR', rev: int = None) -> Path:
    """
    Fetches the event info and skymap from GraceDB

    :param event_name: name of the event
    :param output_dir: directory to save the skymap and event info
    :param rev: revision number of the event
    :return: path to the skymap
    """

    # if Path(event_name).exists():
    #     savepath = Path(event_name)
    # elif output_dir.joinpath(event_name).exists():
    #     savepath = output_dir.joinpath(event_name)
    # elif event_name[:8] == "https://":
    #     savepath = download_from_url(
    #         event_name, output_dir, os.path.basename(event_name)
    #     )
    # else:
    savepath = get_skymap_gracedb(event_name, output_dir=output_dir, rev=rev)

    return savepath



def get_ivorn(xml_file_path):
    try:
        try:
            tree = ET.parse(xml_file_path)
        except:
            print('file not found')
            exit()
        root = tree.getroot()
        ivorn = root.attrib['ivorn']
        # Extracting from the 3rd '/' to the '#' character
        return ivorn.split('/')[3].split('#')[0]
    except Exception as e:
        print("Error:", e)
        return None
    


    
def getEvent(xml_file):
    # Parse the XML file
    tree = ET.parse(xml_file)
    root = tree.getroot()
    
    # Find the skymap_fits parameter within the GW_SKYMAP group
    event = root.find(".//Param[@name='GraceID']").attrib.get('value')
    
    return event

# Description:
#Grabs the GW skymap URL from VOEvent files




def getFERMICoordinates(file_path):
    # Parse the XML file
    tree = ET.parse(file_path)

    # Find the C1, C2, and Error2Radius elements
    c1_element = tree.find(".//C1")
    c2_element = tree.find(".//C2")
    error_radius_element = tree.find(".//Error2Radius")

    if c1_element is not None and c2_element is not None and error_radius_element is not None:
        # Extract RA (C1), Dec (C2), and error radius values
        ra = float(c1_element.text)
        dec = float(c2_element.text)
        error_radius = float(error_radius_element.text)
        return ra, dec, error_radius
    else:
        raise ValueError("<C1>, <C2>, or <Error2Radius> element not found in the XML.")
# Description:
#Grabs FK5 coordinates and error radius from GMB VOEvent.xml files


def decdeg2hms(decimal_degrees):
    mult = -1 if decimal_degrees < 0 else 1
    mnt,sec = divmod(abs(decimal_degrees)*3600, 60)
    deg,mnt = divmod(mnt, 60)
    
    return mult*deg, mult*mnt, mult*sec

# Description:
#Changes decimal degrees to degree, minute, second


def Fermi_fileWrite(output_file_path, radec, fields):
    try:
        with open(output_file_path, 'w') as file:
            i=0
            for row in radec:
                file.write(f"{fields[i]},{row[0]},{row[1]},1\n")
                i+=1
        print(f"Schedule successfully written to {output_file_path}")
    except Exception as e:
        print(f"Error occurred while writing to file: {e}")

# Description:
#makes schedule file for FERMI notices